from concurrent.futures import ThreadPoolExecutor
from threading import Lock
from typing import TYPE_CHECKING, Callable, Dict, List

if TYPE_CHECKING:
    from arkaine.tools.tool import Context, Tool


class Registrar:
    _lock = Lock()
    _enabled = False
    __executor = ThreadPoolExecutor()

    _tools: 'Dict[str, "Tool"]' = {}

    __on_tool_listeners: List[Callable[["Tool"], None]] = []
    __on_tool_call_listeners: List[Callable[["Tool", "Context"], None]] = []

    def __new__(cls):
        raise ValueError("Registrar cannot be instantiated")

    @classmethod
    def register(cls, tool: "Tool"):
        with cls._lock:
            if tool.id in cls._tools:
                pass
            cls._tools[tool.id] = tool

            for listener in cls.__on_tool_listeners:
                cls.__executor.submit(listener, tool)

            tool.add_on_call_listener(cls._on_tool_call)

    @classmethod
    def _on_tool_call(cls, tool: "Tool", ctx: "Context"):
        """
        Whenever a tool we are aware of is called, notify the listener
        """
        with cls._lock:
            if cls._enabled:
                for listener in cls.__on_tool_call_listeners:
                    cls.__executor.submit(listener, tool, ctx)

    @classmethod
    def add_on_tool_register(cls, listener: Callable[["Tool"], None]):
        with cls._lock:
            cls.__on_tool_listeners.append(listener)

    @classmethod
    def get_tools(cls):
        with cls._lock:
            return list(cls._tools.values())

    @classmethod
    def get_tool(cls, identifier: str) -> "Tool":
        with cls._lock:
            for tool in cls._tools.values():
                if tool.id == identifier:
                    return tool
                if tool.name == identifier:
                    return tool
            raise ValueError(f"Tool with identifier {identifier} not found")

    @classmethod
    def add_tool_call_listener(
        cls, listener: Callable[["Tool", "Context"], None]
    ):
        with cls._lock:
            cls.__on_tool_call_listeners.append(listener)

    @classmethod
    def enable(cls):
        with cls._lock:
            cls._enabled = True

    @classmethod
    def disable(cls):
        with cls._lock:
            cls._enabled = False

    @classmethod
    def set_auto_registry(cls, enabled: bool):
        with cls._lock:
            cls._enabled = enabled

    @classmethod
    def is_enabled(cls):
        with cls._lock:
            return cls._enabled



"""Module for managing global context options in a thread-safe manner."""

from pathlib import Path
from threading import Lock
from typing import Optional


class ContextOptions:
    """
    Thread-safe singleton class for managing global context options.
    Provides centralized control over context behavior across the application.
    """

    __instance: Optional["ContextOptions"] = None
    __lock = Lock()

    # Debug enables/disables the debug store on contexts, and whether
    # or not their messages are broadcasted to any possible subscribers
    __debug = False

    __save_on_success = False
    __save_on_exception = False

    # Saving a context will write to the target folder (default, current
    # working directory + "{date}/", the contexts for the current date are
    # saved in the "success" folder. The {date} template will always be
    # replaced with the current day according to the system's timezone. The
    # contexts are then placed in a subfolder which lands it in, by default,
    # the "success" (success_folder) or "exception" (exception_folder) folder.
    # If a context had neither completed or failed, it is saved in the
    # "running" (running_folder) folder.
    __save_folder = Path.cwd() / "{date}"
    __success_folder = "success"
    __running_folder = "running"
    __exception_folder = "exception"

    def __new__(cls):
        raise ValueError("ContextOptions cannot be instantiated")

    @classmethod
    def debug(cls, value: Optional[bool] = None) -> bool:
        with cls.__lock:
            if value is not None:
                cls.__debug = value
            return cls.__debug

    @classmethod
    def save_on_success(cls, value: Optional[bool] = None) -> bool:
        with cls.__lock:
            if value is not None:
                cls.__save_on_success = value
            return cls.__save_on_success

    @classmethod
    def save_on_exception(cls, value: Optional[bool] = None) -> bool:
        with cls.__lock:
            if value is not None:
                cls.__save_on_exception = value
            return cls.__save_on_exception

    @classmethod
    def save_folder(cls, value: Optional[Path] = None) -> Path:
        with cls.__lock:
            if value is not None:
                cls.__save_folder = value
                cls.__ensure_exists(cls.__save_folder)
                cls.__ensure_exists(cls.__success_folder)
                cls.__ensure_exists(cls.__running_folder)
                cls.__ensure_exists(cls.__exception_folder)
            return cls.__save_folder

    @classmethod
    def success_folder(cls, value: Optional[Path] = None) -> Path:
        with cls.__lock:
            if value is not None:
                cls.__success_folder = value
            cls.__ensure_exists(cls.__success_folder)
            return cls.__success_folder

    @classmethod
    def exception_folder(cls, value: Optional[Path] = None) -> Path:
        with cls.__lock:
            if value is not None:
                cls.__exception_folder = value
            cls.__ensure_exists(cls.__exception_folder)
            return cls.__exception_folder

    @classmethod
    def running_folder(cls, value: Optional[Path] = None) -> Path:
        with cls.__lock:
            if value is not None:
                cls.__running_folder = value
            cls.__ensure_exists(cls.__running_folder)
            return cls.__running_folder

    @classmethod
    def __ensure_exists(cls, folder: Path):
        if not folder.exists():
            folder.mkdir(parents=True, exist_ok=True)



from arkaine.tools.argument import Argument, InvalidArgumentException
from arkaine.tools.events import Event, ToolCalled, ToolReturn, ToolStart
from arkaine.tools.example import Example
from arkaine.tools.tool import Context, Tool
from arkaine.tools.toolify import toolify
from arkaine.tools.types import ToolArguments, ToolCalls, ToolResults
from arkaine.tools.wrapper import Wrapper

__all__ = [
    "Argument",
    "InvalidArgumentException",
    "Event",
    "ToolCalled",
    "ToolReturn",
    "ToolStart",
    "Example",
    "Context",
    "Tool",
    "toolify",
]



# from arkaine.llms.claude import Claude
# from arkaine.llms.grow import Groq
# from arkaine.llms.llm import LLM
# from arkaine.llms.ollama import Ollama
# from arkaine.llms.openai import OpenAI

# __all__ = [
#     "Claude",
#     "Groq",
#     "LLM",
#     "Ollama",
#     "OpenAI",
# ]



from __future__ import annotations

import json
import re
from typing import Any, Dict, List, Optional, Tuple


class PromptTemplate:
    """
    PromptTemplate is a class designed to handle easy templating within
    prompts. It can successfully deal with the Prompt type (str or a List of
    dicts) and handle save/load appropriately.
    """

    def __init__(
        self,
        template: str | Dict[str, str],
        template_delimiters: Tuple[str, str] = ("{", "}"),
        defaults: Optional[Dict[str, Any]] = None,
    ):
        self.template = template
        self.template_delimiters = template_delimiters
        self.variables = self.__get_all_variables()
        self.defaults = defaults or {}

    @classmethod
    def from_file(
        cls, path: str, defaults: Optional[Dict[str, Any]] = None
    ) -> PromptTemplate:
        """Load a template from a given filepath."""
        if path.endswith(".json"):
            with open(path, "r") as f:
                template = json.load(f)
                # For JSON files, we want the content of the template
                if isinstance(template, dict):
                    template = next(iter(template.values()))
        else:
            with open(path, "r") as f:
                template = f.read()
        return cls(template, defaults=defaults)

    def __get_all_variables(self) -> Dict[str, Optional[Any]]:
        """
        Run through the template and identify all templating variables.
        """
        if isinstance(self.template, str):
            return {
                var: None
                for var in self.__isolate_templated_variables(self.template)
            }
        elif isinstance(self.template, dict):
            variables: Dict[str, Optional[Any]] = {}
            for content in self.template.values():
                if isinstance(content, str):
                    for var in self.__isolate_templated_variables(content):
                        variables[var] = None
            return variables
        else:
            raise ValueError(
                f"Template must be str or dict, not {type(self.template)}"
            )

    def __isolate_templated_variables(self, text: str) -> List[str]:
        """Find all template variables within text."""
        pattern = (
            rf"\{self.template_delimiters[0]}"
            rf"(\w+)"
            rf"\{self.template_delimiters[1]}"
        )
        return re.findall(pattern, text)

    def __setitem__(self, name: str, value: Any) -> None:
        if name not in self.variables:
            raise ValueError(f"Variable {name} not found in template.")
        self.variables[name] = value

    def __getitem__(self, name: str) -> Any:
        if name not in self.variables:
            raise ValueError(f"Variable {name} not found in template.")
        return self.variables[name]

    def render(
        self, variables: Optional[Dict[str, any]] = None, role: str = "system"
    ) -> List[Dict[str, str]]:
        """Render the template with the given variables."""
        if variables is None:
            variables = self.variables

        # Merge defaults with provided variables, prioritizing provided variables
        merged_variables = self.defaults.copy()
        merged_variables.update(variables)

        template_text = (
            self.template
            if isinstance(self.template, str)
            else next(iter(self.template.values()))
        )

        delimiter_pattern = (
            re.escape(self.template_delimiters[0])
            + r"(.*?)"
            + re.escape(self.template_delimiters[1])
        )

        text = template_text
        for var, value in merged_variables.items():
            pattern = delimiter_pattern.replace("(.*?)", re.escape(var))
            sanitized_value = str(value).replace("\\", "\\\\")
            text = re.sub(pattern, sanitized_value, text)

        return [{"role": role, "content": text}]

    @classmethod
    def default(cls) -> PromptTemplate:
        """Create a default prompt template with agent_explanation and task
        variables."""
        template = "{agent_explanation}\n\n{task}"
        return cls(template)



from __future__ import annotations

import json
import threading
from typing import Dict, Set

from websockets.server import WebSocketServerProtocol
from websockets.sync.server import serve

from arkaine.registrar.registrar import Registrar
from arkaine.tools.events import ToolException, ToolReturn
from arkaine.tools.tool import Context, Event, Tool


class ComposerSocket:
    """
    ComposerSocket handles WebSocket connections and broadcasts context events
    to connected clients.
    """

    def __init__(self, port: int = 9001, max_contexts: int = 1024):
        """
        Initialize a ComposerSocket that creates its own WebSocket endpoint.

        Args:
            port (int): The port to run the WebSocket server on (default: 9001)
            max_contexts (int): The maximum number of contexts to keep in
                memory (default: 1024)
        """
        self.port = port
        self.active_connections: Set[WebSocketServerProtocol] = set()
        self._contexts: Dict[str, Context] = {}
        self._tools: Dict[str, Tool] = {}
        self._server = None
        self._server_thread = None
        self._running = False
        self._lock = threading.Lock()
        self.__max_contexts = max_contexts

        Registrar.enable()
        Registrar.add_tool_call_listener(self._on_tool_call)

        Registrar.add_on_tool_register(self._on_tool_register)

        with self._lock:
            tools = Registrar.get_tools()
            for tool in tools:
                self._tools[tool.id] = tool

    def _on_tool_call(self, tool: Tool, context: Context):
        # Subscribe to all the context's events for this tool from
        # here on out if its a root context
        self._handle_context_creation(context)
        self._broadcast_context(context)
        context.add_event_listener(
            self._broadcast_event, ignore_children_events=True
        )
        context.add_on_end_listener(self._context_complete)

    def _context_complete(self, context: Context):
        if context.exception:
            self._broadcast_event(context, ToolException(context.exception))
        else:
            self._broadcast_event(context, ToolReturn(context.output))

    def _on_tool_register(self, tool: Tool):
        with self._lock:
            self._tools[tool.id] = tool
        self._broadcast_tool(tool)

    def _handle_context_creation(self, context: Context):
        """
        Add the context to the internal state memory and remove contexts by
        age if over a certain threshold.
        """
        with self._lock:
            if not context.is_root:
                return
            self._contexts[context.id] = context
            if len(self._contexts) > self.__max_contexts:
                oldest_context = min(
                    self._contexts.values(), key=lambda x: x.created_at
                )
                del self._contexts[oldest_context.id]

    def _broadcast_to_clients(self, message: dict):
        """Helper function to broadcast a message to all active clients"""
        with self._lock:
            dead_connections = set()
            for websocket in self.active_connections:
                try:
                    websocket.send(json.dumps(message))
                except Exception as e:
                    print(f"Failed to send to client {websocket}: {e}")
                    dead_connections.add(websocket)

            # Clean up dead connections
            self.active_connections -= dead_connections

    def _handle_client(self, websocket):
        """Handle an individual client connection"""
        try:
            remote_addr = websocket.remote_address
            print(f"New client connected from {remote_addr}")
        except Exception:
            remote_addr = "unknown"
            print("New client connected (address unknown)")

        try:
            with self._lock:
                self.active_connections.add(websocket)
                # Send initial context states and their events immediately

                for tool in self._tools.values():
                    try:
                        websocket.send(
                            json.dumps(self.__build_tool_message(tool))
                        )
                    except Exception as e:
                        print(f"Failed to send initial tool state: {e}")
                        return

                for context in self._contexts.values():
                    try:
                        websocket.send(
                            json.dumps(self.__build_context_message(context))
                        )

                    except Exception as e:
                        print(f"Failed to send initial context state: {e}")
                        return

            # Keep connection alive until client disconnects or server stops
            while self._running:
                try:
                    message = websocket.recv(timeout=1)
                    if message:  # Handle any client messages if needed
                        pass
                except TimeoutError:
                    continue
                except Exception:
                    break

        except Exception as e:
            print(f"Client connection error: {e}")
        finally:
            with self._lock:
                self.active_connections.discard(websocket)
            print(f"Client disconnected from {remote_addr}")

    def __build_tool_message(self, tool: Tool):
        return {"type": "tool", "data": tool.to_json()}

    def _broadcast_tool(self, tool: Tool):
        """Broadcast a tool to all active clients"""
        self._broadcast_to_clients(self.__build_tool_message(tool))

    def __build_context_message(self, context: Context):
        return {"type": "context", "data": context.to_json()}

    def _broadcast_context(self, context: Context):
        """Broadcast a context to all active clients"""
        self._broadcast_to_clients(self.__build_context_message(context))

    def _broadcast_event(self, context: Context, event: Event):
        """Broadcasts an event to all active WebSocket connections."""
        event_data = event.to_json()
        self._broadcast_to_clients(
            {
                "type": "event",
                "context_id": context.id,
                "data": event_data,
            }
        )

    def start(self):
        """Start the WebSocket server in a background thread"""
        if self._running:
            return

        self._running = True
        self._server_thread = threading.Thread(
            target=self._run_server, daemon=True
        )
        self._server_thread.start()
        print(f"WebSocket server started on ws://localhost:{self.port}")

    def _run_server(self):
        """Run the WebSocket server"""
        with serve(self._handle_client, "localhost", self.port) as server:
            self._server = server
            server.serve_forever()

    def stop(self):
        """Stop the WebSocket server"""
        self._running = False
        if self._server:
            self._server.shutdown()
        if self._server_thread and self._server_thread.is_alive():
            self._server_thread.join()
        self._server_thread = None
        print("WebSocket server stopped")

    def __del__(self):
        """Clean up resources when the object is deleted"""
        self.stop()



from arkaine.integrations.api import API
from arkaine.integrations.cli import CLI
from arkaine.integrations.rss import RSS, Feed
from arkaine.integrations.schedule import Schedule, Task
from arkaine.utils.interval import Interval

__all__ = ["API", "CLI", "Schedule", "Task", "Interval", "RSS", "Feed"]



import socket
from {client_import} import *

import code
import sys
import os

from contextlib import redirect_stdout, redirect_stderr
from typing import Any, Callable, Dict, List, Tuple
from io import StringIO

class REPL(code.InteractiveInterpreter):
    def __init__(self, variables: Dict[Any, Any] = {}, funcs: List[Callable] = []):
        self.locals = {}
        self.globals = variables
        for func in funcs:
            self.globals[func.__name__] = func
        self.__stdout = StringIO()
        self.__stderr = StringIO()
    
    def runcode(self, code: str) -> Tuple[str, str]:
        """
        Run the given code, storing locals and globals
        over time. Return the stdout and stderr respectively
        of the execution.
        """
        with redirect_stdout(self.__stdout), redirect_stderr(self.__stderr):
            super().runcode(code, self.locals, self.globals)
        return self.__stdout.getvalue(), self.__stderr.getvalue()
    

def __repl_main():
    repl = REPL(
        funcs=[
            {tool_names}
        ]
    )

    sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    sock.connect("/{code_directory}/{socket_file}")
    while True:
        data = __wait_for_data(sock)
        
        if data["function"] == "repl":
            stdout, stderr = repl.runcode(data["code"])
            __send_result(stdout, stderr)


if __name__ == "__main__":
    __repl_main()


import io
from typing import TextIO

import pytest

from arkaine.events import AgentCalled, AgentLLMResponse, AgentPrompt
from arkaine.logging.logger import Colors, Logger
from arkaine.tools.events import ToolCalled, ToolException, ToolReturn
from arkaine.tools.tool import Argument, Context, Tool


class MockTool(Tool):
    def __init__(self, name: str = "mock_tool"):
        super().__init__(
            name=name,
            description="A mock tool for testing",
            args=[
                Argument(
                    name="test_arg",
                    description="A test argument",
                    type="string",
                    required=True,
                )
            ],
            func=lambda: None,
        )


@pytest.fixture
def output_stream() -> TextIO:
    return io.StringIO()


@pytest.fixture
def logger(output_stream: TextIO) -> Logger:
    return Logger(output_stream=output_stream, use_colors=True)


@pytest.fixture
def tool() -> Tool:
    return MockTool()


@pytest.fixture
def context(tool: Tool) -> Context:
    return Context(tool=tool)


def test_logger_initialization():
    """Test logger initializes with default values"""
    logger = Logger()
    assert logger.use_colors is True
    assert logger.indent_size == 2


def test_logger_color_disabled(output_stream: TextIO):
    """Test logger respects color disable flag"""
    logger = Logger(output_stream=output_stream, use_colors=False)
    context = Context(MockTool())

    event = ToolCalled({"arg": "value"})
    logger.log_event(context, event)

    output = output_stream.getvalue()
    assert Colors.RED not in output
    assert Colors.GREEN not in output
    assert Colors.RESET not in output


def test_tool_event_logging(
    logger: Logger, context: Context, output_stream: TextIO
):
    """Test logging of tool-related events"""
    # Test tool called event
    tool_called = ToolCalled({"test_arg": "value"})
    logger.log_event(context, tool_called)

    # Test tool return event
    tool_return = ToolReturn("test result")
    logger.log_event(context, tool_return)

    output = output_stream.getvalue()
    assert "test_arg" in output
    assert "value" in output
    assert "test result" in output


def test_agent_event_logging(
    logger: Logger, context: Context, output_stream: TextIO
):
    """Test logging of agent-related events"""
    # Test agent called event
    agent_called = AgentCalled({"test_arg": "value"})
    logger.log_event(context, agent_called)

    # Test agent prompt event
    agent_prompt = AgentPrompt("test prompt")
    logger.log_event(context, agent_prompt)

    # Test agent LLM response event
    agent_response = AgentLLMResponse("test response")
    logger.log_event(context, agent_response)

    output = output_stream.getvalue()
    assert "test prompt" in output
    assert "test response" in output


def test_exception_logging(
    logger: Logger, context: Context, output_stream: TextIO
):
    """Test logging of exceptions"""
    exception = ValueError("Test error")
    event = ToolException(exception)
    logger.log_event(context, event)

    output = output_stream.getvalue()
    assert "Test error" in output
    assert Colors.RED in output


def test_json_formatting(
    logger: Logger, context: Context, output_stream: TextIO
):
    """Test JSON formatting of complex data structures"""
    data = {"key": "value", "nested": {"inner": "data"}}
    event = ToolReturn(data)
    logger.log_event(context, event)

    output = output_stream.getvalue()
    assert '"key": "value"' in output
    assert '"nested"' in output
    assert '"inner": "data"' in output


def test_tool_attachment(logger: Logger, tool: Tool):
    """Test attaching a tool to the logger"""
    logger.attach_tool(tool)

    # To verify a tool is attached, I should be able to determine
    # that it was called
    context = Context(tool)
    tool_called = ToolCalled({})
    logger.log_event(context, tool_called)

    output = logger.output_stream.getvalue()
    assert tool.id in output


def test_indentation(logger: Logger, context: Context, output_stream: TextIO):
    """Test proper indentation of multi-line output"""
    data = {"key1": "value1", "key2": {"nested": "value2"}}
    event = ToolReturn(data)
    logger.log_event(context, event)

    output = output_stream.getvalue()
    lines = output.split("\n")

    # Make assertion more flexible to handle different indentation styles
    assert any(
        line.strip().startswith('"') and line.startswith(" ") for line in lines
    ), "Expected to find indented JSON lines"


def test_thread_safety(logger: Logger, context: Context):
    """Test thread-safe logging with multiple concurrent events"""
    import threading
    import time

    def log_events():
        for i in range(5):
            event = ToolCalled({"count": i})
            logger.log_event(context, event)
            time.sleep(0.01)

    threads = [threading.Thread(target=log_events) for _ in range(3)]
    for thread in threads:
        thread.start()
    for thread in threads:
        thread.join()

    output = logger.output_stream.getvalue()
    # Verify all events were logged (5 events * 3 threads = 15 events)
    assert output.count("mock_tool") == 15



from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List, Optional

from arkaine.backends.base import BaseBackend
from arkaine.events import AgentLLMCalled, AgentLLMResponse, AgentPrompt
from arkaine.llms.llm import LLM, Prompt
from arkaine.tools.tool import Argument, Context, Example, Tool


class Agent(Tool, ABC):
    def __init__(
        self,
        name: str,
        description: str,
        args: List[Argument],
        llm: LLM,
        examples: List[Example] = [],
        process_answer: Optional[Callable[[str], Any]] = None,
    ):
        """
        An agent is a tool that utilizes an LLM. Prove an LLM model to generate
        completions, and implement prepare_prompt to convert incoming arguments
        to a prompt for your agent.

        The optional process_answer is a function that is fed the raw output of
        the LLM and converted in whatever manner you wish. If it is not
        provided, the raw output of the LLM is simply returned instead.
        """
        super().__init__(name, description, args, None, examples)
        self.llm = llm
        self.process_answer = process_answer

    @abstractmethod
    def prepare_prompt(self, **kwargs) -> Prompt:
        """
        Given the arguments for the agent, create the prompt to feed to the LLM
        for execution.
        """
        pass

    def invoke(self, context: Context, **kwargs) -> Any:
        prompt = self.prepare_prompt(**kwargs)
        context.broadcast(AgentPrompt(prompt))
        context.broadcast(AgentLLMCalled())
        result = self.llm.completion(prompt)
        context.broadcast(AgentLLMResponse(result))

        final_result = (
            self.process_answer(result) if self.process_answer else result
        )

        return final_result


class MetaAgent(Agent):

    def __init__(
        self,
        name: str,
        description: str,
        args: List[Argument],
        llm: LLM,
        examples: List[Example] = [],
        initial_state: Dict[str, Any] = {},
        max_steps: Optional[int] = None,
    ):
        super().__init__(name, description, args, llm, examples)
        self.__initial_state = initial_state
        self.max_steps = max_steps

    @abstractmethod
    def prepare_prompt(self, context: Context, **kwargs) -> Prompt:
        pass

    @abstractmethod
    def extract_result(self, context: Context, output: str) -> Optional[Any]:
        pass

    def __initialize_state(self, context: Context):
        if self.__initial_state:
            state = self.__initial_state.copy()
            for arg in state:
                context[arg] = state[arg]

    def invoke(self, context: Context, **kwargs) -> Any:
        self.__initialize_state(context)
        step = 0

        while True:
            step += 1
            if self.max_steps and step > self.max_steps:
                raise Exception("Max steps reached")

            prompt = self.prepare_prompt(context, **kwargs)
            context.broadcast(AgentPrompt(prompt))
            context.broadcast(AgentLLMCalled())
            result = self.llm.completion(prompt)
            context.broadcast(AgentLLMResponse(result))

            result = self.extract_result(context, result)
            if result is not None:
                return result


class BackendAgent(Tool, ABC):

    def __init__(
        self,
        name: str,
        description: str,
        args: List[Argument],
        backend: BaseBackend,
        examples: List[Example] = [],
    ):
        super().__init__(name, description, args, None, examples)
        self.backend = backend

    @abstractmethod
    def prepare_for_backend(self, **kwargs) -> Dict[str, Any]:
        """
        Given the arguments for the agent, transform them
        (if needed) for the backend's format. These will be
        passed to the backend as arguments.
        """
        pass

    def invoke(self, context: Context, **kwargs) -> Any:
        return self.backend.invoke(context, self.prepare_for_backend(**kwargs))



import math
import re
from typing import List, Optional, Tuple, Union

import numpy as np
import ollama


def isolate_sentences(text: str) -> List[str]:
    sentences = []
    current_sentence = ""
    for word in text.split():
        current_sentence += word.strip() + " "
        if word.endswith(".") or word.endswith("?") or word.endswith("!"):
            sentences.append(current_sentence.strip())
            current_sentence = ""

    return sentences


def chunk_text_by_sentences(
    text: str,
    sentences_per: int,
    overlap: int = 0,
    isolate_paragraphs: bool = False,
):
    if isolate_paragraphs:
        paragraphs = re.split(r"\n{2,}", text)
        text = [p.strip() for p in paragraphs if p.strip()]
    else:
        text = [text]

    chunks = []

    for paragraph in text:
        sentences = isolate_sentences(paragraph)
        while len(sentences) > 0:
            chunks.append(" ".join(sentences[0:sentences_per]))
            sentences = sentences[sentences_per + 1 - overlap :]

    return chunks


def cosine_distance(a: List[float], b: List[float]):
    """
    cosine_distance calculates the cosine distance between two vectors, which
    are assumed to be a 1-d array of floats. Returns 1.0 if either vector is
    all zeros.
    """
    uv = np.dot(a, b)
    uu = np.dot(a, a)
    vv = np.dot(b, b)
    dist = uv / math.sqrt(uu * vv)

    return np.clip(dist, 0.0, 2.0)


class InMemoryEmbeddingStore:
    """
    InMemoryEmbeddingStore holds a set of text chunks in memory with an
    associated embedding vector. Then, on query, it will return the closest
    embeddings in memory to that vector.

    This is to be used for singular documents or quick ephemeral searches,
    where another more permanent store would be too
    """

    def __init__(self, embedding_model: Optional[str] = None):
        if embedding_model is None:
            embedding_model = "all-minilm:latest"

        self.__embedding_model__ = embedding_model
        self.__memory__: List[Tuple[str, List[float]]] = []

    def add_text(self, content: Union[str, List[str]]) -> List[List[float]]:
        if isinstance(content, str):
            content = [content]

        embeddings = []

        for text in content:
            embedding = self.get_embedding(text)
            embeddings.append(embedding)
            self.__memory__.append((text, embedding))

        return embeddings

    def __measure_distance(self, a: List[float], b: List[float]) -> float:
        return cosine_distance(a, b)

    def get_embedding(self, text: str) -> List[float]:
        return ollama.embeddings(model=self.__embedding_model__, prompt=text)[
            "embedding"
        ]

    def query(
        self,
        query: str,
        top_n: Optional[int] = 10,
    ) -> List[str]:
        query_embedding = self.get_embedding(query)

        # TODO - make more efficient
        results: List[Tuple[str, float]] = []

        for item in self.__memory__:
            text, embedding = item

            # Find the cosine distance
            distance = self.__measure_distance(query_embedding, embedding)

            results.append([text, distance])

        # Sort the results based on the distance, highest value is best
        results = sorted(results, key=lambda x: x[1], reverse=True)

        if top_n:
            results = results[:top_n]

        out = [x[0] for x in results]

        return out



from __future__ import annotations

import email
import imaplib
import os
import re
from datetime import datetime
from threading import Lock, Thread
from time import sleep
from typing import Callable, Dict, List, Optional, Union

from arkaine.toolbox.email import EmailSender
from arkaine.tools import Tool
from arkaine.utils.interval import Interval


class EmailMessage:
    """Represents a received email message."""

    def __init__(
        self,
        subject: str,
        sender: str,
        body: str,
        received_at: datetime,
        message_id: str,
    ):
        self.subject = subject
        self.sender = sender
        self.body = body
        self.received_at = received_at
        self.message_id = message_id

    def __str__(self):
        return f"From: {self.sender}\nSubject: {self.subject}\nDate: {self.received_at}"

    @staticmethod
    def from_message(msg: email.message.Message) -> EmailMessage:
        """Create EmailMessage from email.message.Message object."""
        subject = msg.get("subject", "")
        sender = msg.get("from", "")
        received_at = email.utils.parsedate_to_datetime(msg.get("date"))
        message_id = msg.get("message-id", "")

        # Extract body (handle both plain text and HTML)
        body = ""
        if msg.is_multipart():
            for part in msg.walk():
                if part.get_content_type() in ["text/plain", "text/html"]:
                    body = part.get_payload(decode=True).decode()
                    break
        else:
            body = msg.get_payload(decode=True).decode()

        return EmailMessage(
            subject=subject,
            sender=sender,
            body=body,
            received_at=received_at,
            message_id=message_id,
        )


class Inbox:
    """
    Monitors incoming emails and triggers tools when matching emails are found.

    Basic usage:
    ```python
    # Create monitor checking every 5 minutes
    inbox = Inbox(
        username="user@example.com",
        password="pass",
        service="gmail",
        check_every="5:minutes"
    )

    # Add filter and associated tools
    inbox.add_filter(
        EmailFilter(subject_pattern="Important:.*"),
        tools=[notification_tool]
    )

    # Start monitoring
    inbox.start()

    # Add another filter while running
    inbox.add_filter(
        EmailFilter(sender_pattern="boss@company.com"),
        tools=[urgent_tool]
    )

    # Stop monitoring
    inbox.stop()
    ```
    """

    # Common IMAP servers
    COMMON_IMAP_SERVERS = {
        "gmail": "imap.gmail.com",
        "outlook": "outlook.office365.com",
        "yahoo": "imap.mail.yahoo.com",
        "aol": "imap.aol.com",
        "icloud": "imap.mail.me.com",
    }

    def __init__(
        self,
        tools: List[Tool],
        filters: List[Callable[[EmailMessage], bool]],
        username: Optional[Union[str, dict]] = None,
        password: Optional[Union[str, dict]] = None,
        service: Optional[str] = None,
        imap_host: Optional[str] = None,
        check_every: Union[str, Interval] = "5:minutes",
        env_prefix: str = "EMAIL",
    ):
        self.username = self._load_credential(
            username, f"{env_prefix}_USERNAME"
        )
        self.password = self._load_credential(
            password, f"{env_prefix}_PASSWORD"
        )

        # Set up IMAP host
        if imap_host:
            self.imap_host = imap_host
        elif service:
            if service not in self.COMMON_IMAP_SERVERS:
                raise ValueError(f"Unknown email service: {service}")
            self.imap_host = self.COMMON_IMAP_SERVERS[service]
        else:
            self.imap_host = os.getenv(f"{env_prefix}_IMAP_HOST")
            if not self.imap_host:
                raise ValueError("IMAP host not provided")

        # Set up interval
        if isinstance(check_every, str):
            self.interval = Interval(datetime.now(), recur_every=check_every)
        else:
            self.interval = check_every

        self.__tools: List[Tool] = tools
        self.__filters: Dict[EmailFilter, List[Tool]] = {}

        self.seen_messages: set = set()

        self.__running = False
        self.__lock = Lock()
        self.__thread = Thread(target=self.__run)
        self.__thread.daemon = True
        self.__thread.start()

    def _load_credential(
        self,
        value: Optional[Union[str, dict]],
        env_var: str,
        required: bool = True,
    ) -> Optional[str]:
        """Load a credential from various sources."""
        if isinstance(value, str):
            return value
        if isinstance(value, dict) and "env" in value:
            value = os.getenv(value["env"])
        else:
            value = os.getenv(env_var)

        if required and not value:
            raise ValueError(
                f"Required credential not provided: direct value, "
                f"dict with 'env', or environment variable {env_var}"
            )
        return value

    def add_tool(self, tool: Tool):
        """Add a tool to the inbox."""
        with self.__lock:
            self.tools.append(tool)

    def add_filter(
        self, email_filter: EmailFilter, tools: Union[Tool, List[Tool]]
    ):
        """Add a new filter and associated tools."""
        if isinstance(tools, Tool):
            tools = [tools]
        with self.__lock:
            self.filters[email_filter] = tools

    def start(self):
        """Start monitoring emails."""
        with self.__lock:
            self.__running = True

    def stop(self):
        """Stop monitoring emails."""
        with self.__lock:
            self.__running = False

    def __check_emails(self):
        """Check for new emails and process them."""
        try:
            with imaplib.IMAP4_SSL(self.imap_host) as imap:
                imap.login(self.username, self.password)
                imap.select("INBOX")

                # Search for all emails from the last check
                _, message_numbers = imap.search(None, "ALL")

                for num in message_numbers[0].split():
                    # Fetch message data
                    _, msg_data = imap.fetch(num, "(RFC822)")
                    email_body = msg_data[0][1]
                    msg = email.message_from_bytes(email_body)

                    # Create EmailMessage object
                    message = EmailMessage.from_message(msg)

                    # Skip if we've seen this message before
                    if message.message_id in self.seen_messages:
                        continue

                    self.seen_messages.add(message.message_id)

                    # Check against filters and trigger tools
                    for email_filter, tools in self.filters.items():
                        if email_filter.matches(message):
                            for tool in tools:
                                tool.async_call(message)

        except Exception as e:
            print(f"Error checking emails: {str(e)}")

    def __run(self):
        """Main monitoring loop."""
        while True:
            with self.__lock:
                if not self.__running:
                    sleep(1.0)
                    continue

                if self.interval.trigger_at <= datetime.now():
                    self.__check_emails()
                    self.interval.trigger()

            sleep(1.0)



import os
from typing import Optional

from groq import Groq as GroqAPI

from arkaine.agent import Prompt
from arkaine.llms.llm import LLM


class Groq(LLM):

    CONTEXT_LENGTHS = {
        "gemma2-9b-it": 8192,
        "gemma-7b-it": 8192,
        "llama-3.3-70b-versatile": 32768,
        "llama-3.1-70b-versatile": 32768,
        "llama-3.1-8b-instant": 8192,
        "llama-guard-3-8b": 8192,
        "llama3-70b-8192": 8192,
        "llama3-8b-8192": 8192,
        "mixtral-8x7b-32768": 32768,
        "llama3-groq-70b-8192-tool-use-preview": 8192,
        "llama3-groq-8b-8192-tool-use-preview": 8192,
        "llama-3.3-70b-specdec": 8192,
        "llama-3.1-70b-specdec": 8192,
        "llama-3.3-70b-vision-preview": 128768,
        "llama-3.1-70b-vision-preview": 128768,
        "llama-3.2-1b-preview": 1288192,
        "llama-3.2-3b-preview": 1288192,
        "llama-3.2-11b-vision-preview": 1288192,
        "llama-3.2-90b-vision-preview": 128768,
    }

    def __init__(
        self,
        model: str = "llama3-70b-8192",
        api_key: Optional[str] = None,
        context_length: Optional[int] = 8192,
    ):
        if api_key is None:
            api_key = os.environ.get("GROQ_API_KEY")
        self.__client = GroqAPI(api_key=api_key)
        self.__model = model
        if context_length:
            self.__context_length = context_length
        elif model in self.CONTEXT_LENGTHS:
            self.__context_length = self.CONTEXT_LENGTHS[model]
        else:
            raise ValueError(f"Unknown model: {model}")

    @property
    def context_length(self) -> int:
        return self.__context_length

    def completion(self, prompt: Prompt) -> str:
        if isinstance(prompt, str):
            prompt = [
                {
                    "role": "system",
                    "content": prompt,
                }
            ]

        response = self.__client.chat.completions.create(
            model=self.__model,
            messages=prompt,
        )

        return response.choices[0].message.content



import shutil
import tempfile
from typing import Any, Dict, List, Optional, Tuple, Union
from uuid import uuid4

from docker import DockerClient
from docker.models.containers import Container as dContainer


class Volume:
    """
    Volume represents a Docker volume that can be used to persist data
    across container runs. It encapsulates the properties and behaviors
    associated with a Docker volume.

    Attributes:
        name (str): A unique identifier for the volume. If not provided,
            a UUID will be generated.
        remote (str): The path inside the container where the volume will
            be mounted. Default is "/data".
        image (Optional[str]): The Docker image associated with the volume.
        persist_volume (bool): Indicates whether the volume should persist
            after the container is stopped. Default is False.
        read_only (bool): Indicates whether the volume should be mounted as
            read-only. Default is False.

    Methods:
        delete():
            Deletes the volume, even if it is marked as persistent. This method
            is intended for manual cleanup of the volume when persistence is no
            longer needed.
    """

    def __init__(
        self,
        name: Optional[str] = None,
        remote: str = "/data",
        image: Optional[str] = None,
        persist_volume: bool = False,
        read_only: bool = False,
    ):
        if name is None:
            self.__name = f"arkaine-{uuid4()}"
        else:
            self.__name = name
        self.__remote = remote
        self.__image = image
        self.__persist_volume = persist_volume
        self.__read_only = read_only

    @property
    def name(self) -> str:
        return self.__name

    @property
    def remote(self) -> str:
        return self.__remote

    @property
    def read_only(self) -> bool:
        return self.__read_only

    @property
    def image(self) -> str:
        return self.__image

    @property
    def persist_volume(self) -> bool:
        return self.__persist_volume

    def mount_args(self) -> Dict[str, str]:
        return {
            "type": "volume",
            "source": self.__name,
            "target": self.__remote,
            "read_only": self.__read_only,
        }

    def delete(self):
        """
        Deletes the volume, *even if persist volume is set*; this
        is meant as an "overwrite" method to manually clean up the
        volume after its persistence is unneeded.
        """
        self.__persist_volume = False
        del self

    def __del__(self):
        if self.__persist_volume:
            pass
        else:
            self.__client.volumes.get(self.__name).remove()


class BindVolume:
    """
    BindVolume represents a bind mount in Docker, allowing a local directory
    to be mounted into a container. This is useful for sharing files between
    the host and the container.

    Attributes:
        local (Optional[str]): The local path on the host machine to be
            mounted. If not provided, a temporary directory will be created.
        remote (str): The path inside the container where the bind mount will
            be located. Default is "/data".
        name (Optional[str]): A unique identifier for the bind volume. If not
            provided, a UUID will be generated.
        read_only (bool): Indicates whether the bind mount should be read-only.
            Default is False.
    """

    def __init__(
        self,
        local: Optional[str] = None,
        remote: str = "/data",
        name: Optional[str] = None,
        read_only: bool = False,
    ):
        if name is None:
            self.__name = f"arkaine-{uuid4()}"
        else:
            self.__name = name
        if local is None:
            self.__local = tempfile.mkdtemp()
            self.__tmpdir = True
        else:
            self.__local = local
            self.__tmpdir = False
        self.__remote = remote
        self.__read_only = read_only

    @property
    def name(self) -> str:
        return self.__name

    @property
    def local(self) -> str:
        return self.__local

    @property
    def remote(self) -> str:
        return self.__remote

    @property
    def read_only(self) -> bool:
        return self.__read_only

    @property
    def image(self) -> str:
        return self.__image

    @property
    def persist_volume(self) -> bool:
        return self.__persist_volume

    def move_to(self, from_path: str, to_path: str):
        pass

    def mount_args(self) -> Dict[str, str]:
        return {
            "type": "bind",
            "source": self.__local,
            "target": self.__remote,
            "read_only": self.__read_only,
        }

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def __del__(self):
        if self.__tmpdir:
            shutil.rmtree(self.__local)


class Container:
    """
    Container represents a Docker container that can be created, started,
    and managed. It encapsulates the properties and behaviors associated
    with a Docker container.

    Attributes:
        name (str): A unique identifier for the container.
        image (str): The Docker image to use for the container. Default is
                     "alpine:latest".
        args (Dict[str, Any]): Arguments to pass to the container at runtime.
        env (Dict[str, Any]): Environment variables to set in the container.
        volumes (List[Union[BindVolume, Volume]]): A list of volumes to mount
            in the container.
        ports (Dict[str, str]): Port mappings between the container and the
            host.
        entrypoint (Optional[str]): The command to run when the container
            starts.
        command (Optional[str]): The command to execute in the container.

    Methods:
        start():
            Starts the container with the specified configuration, running a
            command to keep it alive.

        run(command: Optional[str]) -> Tuple[str, str]:
            Executes a command in the running container and returns the output
            and error messages.

        bash(command: str) -> str:
            Executes a bash command in the container.

        stop():
            Stops and removes the container.
    """

    def __init__(
        self,
        name: str,
        image: str = "alpine:latest",
        args: Dict[str, Any] = {},
        env: Dict[str, Any] = {},
        volumes: List[Union[BindVolume, Volume]] = [],
        ports: Dict[str, str] = [],
        entrypoint: str = None,
        command: Optional[str] = None,
    ):

        self.__name = name
        self.__image = image
        self.__args = args
        self.__env = env
        self.__volumes = volumes

        port_bindings = {}
        if ports:
            for container_port, host_port in ports.items():
                if "/" not in container_port:
                    container_port = f"{container_port}/tcp"
                port_bindings[container_port] = host_port
        self.__ports = port_bindings

        self.__entrypoint = entrypoint
        if command is None:
            self.__command = "sleep infinity"
        else:
            self.__command = command

        self._container: Optional[dContainer] = None
        self.__client = DockerClient.from_env()

    def __image_check(self):
        # Check to see if we have the image; if not, attempt
        # to pull it
        try:
            self.__client.images.get(self.__image)
        except:  # noqa: E722
            try:
                self.__client.images.pull(self.__image)
            except:  # noqa: E722
                pass

    def __wait_for_container(self):
        self._container.reload()
        if self._container.status != "running":
            raise DockerExecutionException(
                f"Container failed to start. Status: {self._container.status}"
            )

    def start(self):
        """
        start runs the container with a "sleep infinity" command to keep
        it alive and running until it is told to stop.
        """
        if self._container:
            return

        self.__image_check()

        self._container = self.__client.containers.run(
            self.__image,
            name=self.__name,
            command="sleep infinity",
            detach=True,
            mounts=[v.mount_args() for v in self.__volumes],
            ports=self.__ports,
            environment=self.__env,
            entrypoint=self.__entrypoint,
        )

        self.__wait_for_container()

    def run(self, command: Optional[str]) -> Tuple[str, str]:
        self.__image_check()

        command = self.__command if command is None else command

        if not self._container:
            self.start()

        # Execute execute execute!
        result = self._container.exec_run(
            command,
            stderr=True,  # Enable stderr capture
            demux=True,  # Split stdout/stderr apart
        )

        # Capture output
        stdout, stderr = result.output
        stdout = stdout.decode("utf-8") if stdout else ""
        stderr = stderr.decode("utf-8") if stderr else ""

        # Check for errors
        if result.exit_code != 0:
            raise DockerExecutionException(stdout, stderr)

        return stdout, stderr

    def bash(self, command: str) -> str:
        # Make sure escape characters are handled for quotations
        command = command.replace("'", "\\'")
        return self.run(f"/bin/bash -c '{command}'")

    def stop(self):
        if self._container:
            self._container.remove(force=True)
            self._container = None

    @property
    def container(self) -> dContainer:
        return self._container

    def __enter__(self):
        self.run(None)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.cleanup()

    def __del__(self):
        self.__exit__(None, None, None)
        self.__client.close()


class DockerExecutionException(Exception):
    def __init__(self, message: str, stderr_or_trace: str = None):
        self.message = message
        self.stderr_or_trace = stderr_or_trace
        super().__init__(self.message)

    def __str__(self):
        if self.stderr_or_trace:
            return f"Python code execution failed:\n{self.message}\n\nError output:\n{self.stderr_or_trace}"
        return f"Python code execution failed: {self.message}"



from arkaine.tools.tool import Tool


def python(
    tool: Tool,
    output_style: str = "standard",
    include_examples: bool = False,
    include_return: bool = True,
) -> str:
    """
    Generate a Python docstring for the given tool.

    Args:
        tool (Tool): The tool for which to generate the docstring.

        output_style (str): The style of output to request (e.g., "standard",
            "google", "numpy").

        include_examples (bool): Whether to include examples in the docstring.

        include_return (bool): Whether to include the return type in the
            docstring (if a Result is specified for the tool).

    Returns:
        str: The generated docstring.
    """
    # Build function signature with tool name and args
    func_sig = f"\ndef {tool.name}("
    if tool.args:
        func_sig += ", ".join(arg.name for arg in tool.args)
    func_sig += ")\n"

    docstring = func_sig + f'"""\n{tool.description}\n\n'

    # Add arguments based on the output style
    if tool.args:
        if output_style == "google":
            docstring += "Args:\n"
            for arg in tool.args:
                desc = arg.description if hasattr(arg, "description") else ""
                arg_desc = f"    {arg.name} ({arg.type}): {desc}\n"
                docstring += arg_desc
        elif output_style == "numpy":
            docstring += "Parameters\n----------\n"
            for arg in tool.args:
                desc = arg.description if hasattr(arg, "description") else ""
                arg_desc = f"{arg.name} : {arg.type}\n    {desc}\n"
                docstring += arg_desc
        elif output_style == "standard":
            docstring += "Args:\n"
            for arg in tool.args:
                desc = arg.description if hasattr(arg, "description") else ""
                arg_desc = f"    {arg.name} ({arg.type}): {desc}\n"
                docstring += arg_desc
        else:
            raise ValueError(f"Invalid output style: {output_style}")

    if tool.result and include_return:
        docstring += f"\n\nReturns: {tool.result}\n"

    if include_examples and tool.examples:
        docstring += "\nExamples:\n"
        for example in tool.examples:
            docstring += f"    {example}\n"

    docstring += '"""'
    return docstring



from typing import Any, List, Optional, Union

import pytest

from arkaine.tools.tool import Tool
from arkaine.tools.toolify import toolify


class TestBasicFormats:
    def test_rst_format(self):
        @toolify
        def func(a: int, b: int) -> int:
            """
            Adds two numbers together

            :param a: The first number to add
            :param b: The second number to add
            :return: The sum of the two numbers
            """
            return a + b

        assert func.name == "func"
        assert "Adds two numbers together" in func.description
        assert "The sum of the two numbers" in func.description
        assert len(func.args) == 2
        assert func.args[0].name == "a"
        assert func.args[0].description == "The first number to add"
        assert func.args[0].type == "int"
        assert func.args[0].required is True

    def test_google_format(self):
        @toolify
        def func(text: str, times: int = 1) -> str:
            """
            Repeats text a specified number of times.

            Args:
                text: The text to repeat
                times: Number of times to repeat the text

            Returns:
                The repeated text
            """
            return text * times

        assert len(func.args) == 2
        assert func.args[0].name == "text"
        assert func.args[0].required is True
        assert func.args[1].name == "times"
        assert func.args[1].required is False
        assert func.args[1].default == "1"
        assert "The repeated text" in func.description

    def test_plain_format(self):
        @toolify
        def func(name: str, age: Optional[int] = None) -> str:
            """
            Formats a greeting for a person.

            name -- The person's name
            age -- The person's age (optional)
            returns -- A formatted greeting
            """
            return f"Hello {name}!"

        assert len(func.args) == 2
        assert func.args[0].name == "name"
        assert func.args[0].description == "The person's name"
        assert func.args[1].name == "age"
        assert func.args[1].type == "Optional[int]"
        assert "A formatted greeting" in func.description


class TestComplexCases:
    def test_multiline_parameters(self):
        @toolify
        def func(text: str, width: int = 80) -> str:
            """
            Centers text.

            Args:
                text: The text to center
                      Can be any string you want
                      Multiple lines are fine
                width: The total width
                       Must be greater than text length
                       Defaults to 80

            Returns:
                The centered text
                With proper width
            """
            return text.center(width)

        assert func.args[0].description == (
            "The text to center Can be any string you want Multiple lines are fine"
        )
        assert "The centered text With proper width" in func.description

    def test_mixed_formats(self):
        @toolify
        def func(a: List[int], b: Union[int, float]) -> float:
            """
            Mixed format test.

            :param a: A list of numbers
            Args:
                b: A number to multiply by

            returns -- First return
            :return: Second return
            Returns:
                Third return
            """
            return sum(a) * b

        assert len(func.args) == 2
        assert func.args[0].type == "List[int]"
        assert func.args[1].type == "Union[int, float]"
        # Should use the first return description encountered
        assert "First return" in func.description

    def test_unicode_handling(self):
        @toolify
        def func(text: str) -> str:
            """
            Unicode test.

            Parameters:
                text: Text with unicode — including em-dash
                      and other™ special® characters…

            Returns: ☺ → ♠ ♣ ♥ ♦ ∞ ≠ ≈
            """
            return text

        assert "em-dash" in func.args[0].description
        assert "☺ → ♠" in func.description


class TestEdgeCases:
    def test_empty_docstring(self):
        @toolify
        def func() -> None:
            """"""
            pass

        assert func.description == "Tool for func"
        assert len(func.args) == 0

    def test_no_docstring(self):
        @toolify
        def func():
            pass

        assert func.description == "Tool for func"
        assert len(func.args) == 0

    def test_malformed_rst(self):
        @toolify
        def func(x: int, y: int) -> int:
            """
            Bad RST.

            :param: missing name
            :param x: good parameter
            :param y missing colon
            :return missing colon
            """
            return x + y

        assert len(func.args) == 2
        assert func.args[0].name == "x"
        assert func.args[0].description == "good parameter"
        assert func.args[1].description == f"Parameter y"

    def test_incomplete_google(self):
        @toolify
        def func(x: int, y: int) -> int:
            """
            Bad Google format.

            Args:
                x: This is fine
                y with no colon
                z: (This parameter doesn't exist)

            Returns
                No colon here
            """
            return x + y

        assert len(func.args) == 2
        assert func.args[0].description == "This is fine"
        assert "No colon here" in func.description

    def test_return_in_middle(self):
        @toolify
        def func(x: int, y: int) -> int:
            """
            Return in middle.

            :param x: First number
            :return: The sum
            :param y: Second number
            """
            return x + y

        assert len(func.args) == 2
        assert "The sum" in func.description
        assert func.args[1].name == "y"

    def test_duplicate_sections(self):
        @toolify
        def func(x: str) -> str:
            """
            Duplicate sections.

            Args:
                x: First definition
            Returns:
                First return

            :param x: Second definition
            :return: Second return

            Parameters:
                x: Third definition
            Returns:
                Third return
            """
            return x

        # Should use first definitions encountered
        assert "First definition" in func.args[0].description
        assert "First return" in func.description


class TestToolifyDecorator:
    def test_with_parentheses(self):
        @toolify()
        def func():
            pass

        assert isinstance(func, Tool)

    def test_without_parentheses(self):
        @toolify
        def func():
            pass

        assert isinstance(func, Tool)

    def test_with_name_override(self):
        @toolify(tool_name="custom_name")
        def func():
            pass

        assert func.name == "custom_name"

    def test_with_description_override(self):
        @toolify(tool_description="custom description")
        def func():
            """Original description"""
            pass

        assert func.description == "custom description"


class TestTypeHandling:
    def test_basic_types(self):
        @toolify
        def func(
            a: int,
            b: str,
            c: float,
            d: bool,
            e: bytes,
        ) -> None:
            pass

        types = [arg.type for arg in func.args]
        assert types == ["int", "str", "float", "bool", "bytes"]

    def test_complex_types(self):
        @toolify
        def func(
            a: List[int],
            b: Optional[str],
            c: Union[int, float],
            d: dict[str, Any],
        ) -> None:
            pass

        types = [arg.type for arg in func.args]
        assert "List[int]" in types
        assert "Optional[str]" in types
        assert "Union[int, float]" in types
        assert "dict" in types



class Result:

    def __init__(
        self,
        type: str,
        description: str,
    ):
        self.type = type
        self.description = description

    def __str__(self):
        return f"({self.type}) - {self.description}"
